@@ -2,29 +2,32 @@
 //
 // This can be expensive since we have to walk the append vecs being cleaned up.
 
-use crate::{
-    bank::{Bank, BankSlotDelta, DropCallback},
-    bank_forks::{BankForks, SnapshotConfig},
-    snapshot_package::AccountsPackageSender,
-    snapshot_utils,
-};
-use crossbeam_channel::{Receiver, SendError, Sender};
-use log::*;
-use rand::{thread_rng, Rng};
-use solana_measure::measure::Measure;
-use solana_sdk::{
-    clock::{BankId, Slot},
-    hash::Hash,
-};
-use std::{
-    boxed::Box,
-    fmt::{Debug, Formatter},
-    sync::{
-        atomic::{AtomicBool, Ordering},
-        Arc, RwLock,
+use {
+    crate::{
+        bank::{Bank, BankSlotDelta, DropCallback},
+        bank_forks::BankForks,
+        snapshot_config::SnapshotConfig,
+        snapshot_package::{AccountsPackageSender, SnapshotType},
+        snapshot_utils::{self, SnapshotError},
+    },
+    crossbeam_channel::{Receiver, SendError, Sender},
+    log::*,
+    rand::{thread_rng, Rng},
+    solana_measure::measure::Measure,
+    solana_sdk::{
+        clock::{BankId, Slot},
+        hash::Hash,
+    },
+    std::{
+        boxed::Box,
+        fmt::{Debug, Formatter},
+        sync::{
+            atomic::{AtomicBool, Ordering},
+            Arc, RwLock,
+        },
+        thread::{self, sleep, Builder, JoinHandle},
+        time::{Duration, Instant},
     },
-    thread::{self, sleep, Builder, JoinHandle},
-    time::{Duration, Instant},
 };
 
 const INTERVAL_MS: u64 = 100;
@@ -93,7 +96,8 @@ impl SnapshotRequestHandler {
         test_hash_calculation: bool,
         use_index_hash_calculation: bool,
         non_snapshot_time_us: u128,
-    ) -> Option<u64> {
+        last_full_snapshot_slot: &mut Option<Slot>,
+    ) -> Option<Result<u64, SnapshotError>> {
         self.snapshot_request_receiver
             .try_iter()
             .last()
@@ -107,7 +111,7 @@ impl SnapshotRequestHandler {
                 let previous_hash = if test_hash_calculation {
                     // We have to use the index version here.
                     // We cannot calculate the non-index way because cache has not been flushed and stores don't match reality.
-                    snapshot_root_bank.update_accounts_hash_with_index_option(true, false, None)
+                    snapshot_root_bank.update_accounts_hash_with_index_option(true, false, None, false)
                 } else {
                     Hash::default()
                 };
@@ -146,6 +150,7 @@ impl SnapshotRequestHandler {
                     use_index_hash_calculation,
                     test_hash_calculation,
                     Some(snapshot_root_bank.epoch_schedule().slots_per_epoch),
+                    false,
                 );
                 let hash_for_testing = if test_hash_calculation {
                     assert_eq!(previous_hash, this_hash);
@@ -160,7 +165,7 @@ impl SnapshotRequestHandler {
                 // accounts that were included in the bank delta hash when the bank was frozen,
                 // and if we clean them here, the newly created snapshot's hash may not match
                 // the frozen hash.
-                snapshot_root_bank.clean_accounts(true, false);
+                snapshot_root_bank.clean_accounts(true, false, *last_full_snapshot_slot);
                 clean_time.stop();
 
                 if accounts_db_caching_enabled {
@@ -169,36 +174,62 @@ impl SnapshotRequestHandler {
                     shrink_time.stop();
                 }
 
-                // Generate an accounts package
+                let block_height = snapshot_root_bank.block_height();
+                let snapshot_type = if snapshot_utils::should_take_full_snapshot(
+                    block_height,
+                    self.snapshot_config.full_snapshot_archive_interval_slots,
+                ) {
+                    *last_full_snapshot_slot = Some(snapshot_root_bank.slot());
+                    Some(SnapshotType::FullSnapshot)
+                } else if snapshot_utils::should_take_incremental_snapshot(
+                    block_height,
+                    self.snapshot_config
+                        .incremental_snapshot_archive_interval_slots,
+                    *last_full_snapshot_slot,
+                ) {
+                    Some(SnapshotType::IncrementalSnapshot(
+                        last_full_snapshot_slot.unwrap(),
+                    ))
+                } else {
+                    None
+                };
+
+                // Snapshot the bank and send over an accounts package
                 let mut snapshot_time = Measure::start("snapshot_time");
-                let r = snapshot_utils::snapshot_bank(
+                let result = snapshot_utils::snapshot_bank(
                     &snapshot_root_bank,
                     status_cache_slot_deltas,
                     &self.accounts_package_sender,
-                    &self.snapshot_config.snapshot_path,
-                    &self.snapshot_config.snapshot_package_output_path,
+                    &self.snapshot_config.bank_snapshots_dir,
+                    &self.snapshot_config.snapshot_archives_dir,
                     self.snapshot_config.snapshot_version,
-                    &self.snapshot_config.archive_format,
+                    self.snapshot_config.archive_format,
                     hash_for_testing,
+                    snapshot_type,
                 );
-                if r.is_err() {
+                if let Err(e) = result {
                     warn!(
-                        "Error generating snapshot for bank: {}, err: {:?}",
+                        "Error taking bank snapshot. slot: {}, snapshot type: {:?}, err: {:?}",
                         snapshot_root_bank.slot(),
-                        r
+                        snapshot_type,
+                        e,
                     );
+
+                    if Self::is_snapshot_error_fatal(&e) {
+                        return Err(e);
+                    }
                 }
                 snapshot_time.stop();
-                info!(
-                    "Took bank snapshot. slot: {}, accounts hash: {}, bank hash: {}",
-                    snapshot_root_bank.slot(),
-                    snapshot_root_bank.get_accounts_hash(),
-                    snapshot_root_bank.hash(),
-                );
+                info!("Took bank snapshot. snapshot type: {:?}, slot: {}, accounts hash: {}, bank hash: {}",
+                      snapshot_type,
+                      snapshot_root_bank.slot(),
+                      snapshot_root_bank.get_accounts_hash(),
+                      snapshot_root_bank.hash(),
+                  );
 
                 // Cleanup outdated snapshots
                 let mut purge_old_snapshots_time = Measure::start("purge_old_snapshots_time");
-                snapshot_utils::purge_old_snapshots(&self.snapshot_config.snapshot_path);
+                snapshot_utils::purge_old_bank_snapshots(&self.snapshot_config.bank_snapshots_dir);
                 purge_old_snapshots_time.stop();
                 total_time.stop();
 
@@ -221,9 +252,34 @@ impl SnapshotRequestHandler {
                     ("total_us", total_time.as_us(), i64),
                     ("non_snapshot_time_us", non_snapshot_time_us, i64),
                 );
-                snapshot_root_bank.block_height()
+                Ok(snapshot_root_bank.block_height())
             })
     }
+
+    /// Check if a SnapshotError should be treated as 'fatal' by SnapshotRequestHandler, and
+    /// `handle_snapshot_requests()` in particular.  Fatal errors will cause the node to shutdown.
+    /// Non-fatal errors are logged and then swallowed.
+    ///
+    /// All `SnapshotError`s are enumerated, and there is **NO** default case.  This way, if
+    /// a new error is added to SnapshotError, a conscious decision must be made on how it should
+    /// be handled.
+    fn is_snapshot_error_fatal(err: &SnapshotError) -> bool {
+        match err {
+            SnapshotError::Io(..) => true,
+            SnapshotError::Serialize(..) => true,
+            SnapshotError::ArchiveGenerationFailure(..) => true,
+            SnapshotError::StoragePathSymlinkInvalid => true,
+            SnapshotError::UnpackError(..) => true,
+            SnapshotError::AccountsPackageSendError(..) => true,
+            SnapshotError::IoWithSource(..) => true,
+            SnapshotError::PathToFileNameError(..) => true,
+            SnapshotError::FileNameToStrError(..) => true,
+            SnapshotError::ParseSnapshotArchiveFileNameError(..) => true,
+            SnapshotError::MismatchedBaseSlot(..) => true,
+            SnapshotError::NoSnapshotArchives => true,
+            SnapshotError::MismatchedSlotHash(..) => true,
+        }
+    }
 }
 
 #[derive(Default)]
@@ -267,7 +323,8 @@ impl AbsRequestHandler {
         test_hash_calculation: bool,
         use_index_hash_calculation: bool,
         non_snapshot_time_us: u128,
-    ) -> Option<u64> {
+        last_full_snapshot_slot: &mut Option<Slot>,
+    ) -> Option<Result<u64, SnapshotError>> {
         self.snapshot_request_handler
             .as_ref()
             .and_then(|snapshot_request_handler| {
@@ -276,6 +333,7 @@ impl AbsRequestHandler {
                     test_hash_calculation,
                     use_index_hash_calculation,
                     non_snapshot_time_us,
+                    last_full_snapshot_slot,
                 )
             })
     }
@@ -306,6 +364,7 @@ impl AccountsBackgroundService {
         accounts_db_caching_enabled: bool,
         test_hash_calculation: bool,
         use_index_hash_calculation: bool,
+        mut last_full_snapshot_slot: Option<Slot>,
     ) -> Self {
         info!("AccountsBackgroundService active");
         let exit = exit.clone();
@@ -359,13 +418,15 @@ impl AccountsBackgroundService {
                     // request for `N` to the snapshot request channel before setting a root `R > N`, and
                     // snapshot_request_handler.handle_requests() will always look for the latest
                     // available snapshot in the channel.
-                    let snapshot_block_height = request_handler.handle_snapshot_requests(
-                        accounts_db_caching_enabled,
-                        test_hash_calculation,
-                        use_index_hash_calculation,
-                        non_snapshot_time,
-                    );
-                    if snapshot_block_height.is_some() {
+                    let snapshot_block_height_option_result = request_handler
+                        .handle_snapshot_requests(
+                            accounts_db_caching_enabled,
+                            test_hash_calculation,
+                            use_index_hash_calculation,
+                            non_snapshot_time,
+                            &mut last_full_snapshot_slot,
+                        );
+                    if snapshot_block_height_option_result.is_some() {
                         last_snapshot_end_time = Some(Instant::now());
                     }
 
@@ -377,10 +438,16 @@ impl AccountsBackgroundService {
                         bank.flush_accounts_cache_if_needed();
                     }
 
-                    if let Some(snapshot_block_height) = snapshot_block_height {
+                    if let Some(snapshot_block_height_result) = snapshot_block_height_option_result
+                    {
                         // Safe, see proof above
-                        assert!(last_cleaned_block_height <= snapshot_block_height);
-                        last_cleaned_block_height = snapshot_block_height;
+                        if let Ok(snapshot_block_height) = snapshot_block_height_result {
+                            assert!(last_cleaned_block_height <= snapshot_block_height);
+                            last_cleaned_block_height = snapshot_block_height;
+                        } else {
+                            exit.store(true, Ordering::Relaxed);
+                            return;
+                        }
                     } else {
                         if accounts_db_caching_enabled {
                             bank.shrink_candidate_slots();
@@ -405,7 +472,7 @@ impl AccountsBackgroundService {
                                 // slots >= bank.slot()
                                 bank.force_flush_accounts_cache();
                             }
-                            bank.clean_accounts(true, false);
+                            bank.clean_accounts(true, false, last_full_snapshot_slot);
                             last_cleaned_block_height = bank.block_height();
                         }
                     }
@@ -455,15 +522,17 @@ impl AccountsBackgroundService {
 
 #[cfg(test)]
 mod test {
-    use super::*;
-    use crate::genesis_utils::create_genesis_config;
-    use crossbeam_channel::unbounded;
-    use solana_sdk::{account::AccountSharedData, pubkey::Pubkey};
+    use {
+        super::*,
+        crate::genesis_utils::create_genesis_config,
+        crossbeam_channel::unbounded,
+        solana_sdk::{account::AccountSharedData, pubkey::Pubkey},
+    };
 
     #[test]
     fn test_accounts_background_service_remove_dead_slots() {
         let genesis = create_genesis_config(10);
-        let bank0 = Arc::new(Bank::new(&genesis.genesis_config));
+        let bank0 = Arc::new(Bank::new_for_tests(&genesis.genesis_config));
         let (pruned_banks_sender, pruned_banks_receiver) = unbounded();
         let request_handler = AbsRequestHandler {
             snapshot_request_handler: None,
